from http import HTTPStatus
from typing import AsyncGenerator, Type

import instructor
from openai import AsyncOpenAI
from openai.types.chat import ChatCompletion, ChatCompletionChunk

from ait.core.domain.errors import LLMAdapterError
from ait.core.domain.interfaces import CompletionResponse, T
from ait.core.ports import LLMPort


class InstructorAdapter(LLMPort):
    """
    Instructor implementation of the LLM port.
    """

    def __init__(
        self,
        model: str,
        embedding_model: str,
        api_key: str,
        base_url: str | None = None,
    ):
        self._model = model
        self._embedding_model = embedding_model

        client_kwargs = dict(
            api_key=api_key,
        )
        if not base_url:
            client_kwargs["base_url"] = "http://localhost:11434/v1"
        self.openai_client = AsyncOpenAI(**client_kwargs)  # type: ignore
        self.client = instructor.from_openai(
            client=self.openai_client,
            mode=instructor.Mode.JSON,
        )

    async def chat(self, messages: list[dict[str, str]]) -> CompletionResponse:
        """
        Sends a message to the LLM and returns a structured response.

        Args:
            messages (list[dict[str, str]]): The messages to generate a response to

        Returns:
            str: The response from the LLM
        """
        output: ChatCompletion = await self.openai_client.chat.completions.create(
            model=self._model,
            messages=messages,
            stream=False,
        )
        response = output.choices[0].message.content
        if not response:
            raise LLMAdapterError(
                status_code=HTTPStatus.SERVICE_UNAVAILABLE.value,
                message="No response from the model",
            )
        return CompletionResponse(
            completion=output,
            content=response,
        )

    async def stream(
        self, messages: list[dict[str, str]]
    ) -> AsyncGenerator[CompletionResponse, None]:
        """
        Streams text outputs from the model.

        Args:
            messages (list[dict[str, str]]): The messages to generate a response to

        Returns:
            AsyncGenerator[str, None]: The response from the LLM
        """
        output: AsyncGenerator[
            ChatCompletionChunk, None
        ] = await self.openai_client.chat.completions.create(
            model=self._model,
            messages=messages,
            stream=True,
        )
        async for chunk in output:
            response = chunk.choices[0].delta.content
            if not response:
                continue
            yield CompletionResponse(
                completion=chunk,
                content=response,
            )

    async def asend(
        self,
        messages: list[dict[str, str]],
        response_model: Type[T],
    ) -> CompletionResponse[T]:
        """
        Sends a message to the LLM asynchronously and returns a structured response.

        Args:
            messages (list[dict[str, str]]): The messages to generate a response to
            response_model (Type[T]): The model to return the response as

        Returns:
            CompletionResponse[T]: The response from the LLM
        """

        (
            instance,
            completion,
        ) = await self.client.chat.completions.create_with_completion(
            response_model=response_model,
            model=self._model,
            messages=messages,
        )
        if not instance:
            raise LLMAdapterError(
                status_code=HTTPStatus.SERVICE_UNAVAILABLE.value,
                message="No response content from the model",
            )
        return CompletionResponse(
            completion=completion,
            content=instance,
        )



from abc import ABC, abstractmethod
from typing import Any, Type, TypeVar, Union

from grafo import Node
from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)
S = TypeVar("S", bound=BaseModel)
V = TypeVar("V", bound=BaseModel)


class WorkflowPort(ABC):
    """
    Abstract base class for workflow operations.
    """

    @abstractmethod
    async def task(
        self, path: str, response_model: Type[S] | None = None, **kwargs: Any
    ) -> Union[str, S]:
        """
        Execute a task.

        Args:
            path (str): The path to the prompt template file
            response_model (Type[S] | None): The response model to return the response as
            **kwargs: Additional arguments to pass to the prompt formatter

        Returns:
            Union[str, S]: The response from the LLM with text content
        """
        pass

    @abstractmethod
    async def redirect(
        self,
        source_node: Node[S],
        validation_node: Node[V],
        target_nodes: list[Node[T]] | None = None,
    ) -> None:
        """
        Redirect the workflow.

        Args:
            source_node (Node[S]): The source node.
            validation_node (Node[V]): The validation node.
            target_nodes (list[Node[T]], optional): The target nodes. Defaults to None.
        """
        pass

    @abstractmethod
    async def run(self, *args: Any, **kwargs: Any) -> Any:
        """
        Run the workflow.
        """
        pass



from typing import Literal

import pytest
from pydantic import BaseModel

from ait import AITools, BaseWorkflow, Node, TreeExecutor
from ait.core.domain.errors import BaseError


class FruitPurchase(BaseModel):
    product: Literal["apple", "banana", "orange"]
    quantity: int


class ValidationResult(BaseModel):
    is_valid: bool
    reason: str
    humanized_failure: str | None = None


class MockWorkflowError(BaseError):
    pass


class MockWorkflow(BaseWorkflow):
    def __init__(self, ait: AITools):
        super().__init__(ait, MockWorkflowError)

    async def fruit_purchase(self, **_) -> FruitPurchase:
        return FruitPurchase(product="apple", quantity=5)

    async def fruit_validation(self, **_) -> ValidationResult:
        return ValidationResult(
            is_valid=True,
            reason="The identified purchase matches the user's request.",
            humanized_failure=None,
        )

    async def run(self, message: str) -> FruitPurchase:
        purchase_node = Node[FruitPurchase](
            uuid="purchase_node",
            coroutine=self.fruit_purchase,
            kwargs=dict(
                path="./tests/purchase.md",
                response_model=FruitPurchase,
                message=message,
            ),
        )

        validation_node = Node[ValidationResult](
            uuid="validation_node",
            coroutine=self.fruit_validation,
            kwargs=dict(
                path="./tests/validation.md",
                response_model=ValidationResult,
                message=message,
                purchase=lambda: purchase_node.output,
            ),
        )

        await purchase_node.connect(validation_node)
        executor = TreeExecutor(uuid="Test Workflow", roots=[purchase_node])
        await executor.run()

        if (
            not purchase_node.output
            or not validation_node.output
            or not validation_node.output.is_valid
        ):
            raise ValueError("Purchase validation failed")

        print(purchase_node.output.model_dump_json(indent=4))
        print(validation_node.output.model_dump_json(indent=4))

        return purchase_node.output


@pytest.mark.asyncio
async def test_mock_workflow():
    ait = AITools("qwen3:8b")
    workflow = MockWorkflow(ait)
    result = await workflow.run("I want to buy 5 apples")
    assert isinstance(result, FruitPurchase)
    assert result.product == "apple"
    assert result.quantity == 5


if __name__ == "__main__":
    pytest.main([__file__])



from abc import ABC, abstractmethod
from typing import Any, Type, TypeVar

from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


class ModellerPort(ABC):
    """
    Abstract base class for model operations.
    """

    @abstractmethod
    def inject_types(
        self,
        model: Type[T],
        fields: list[tuple[str, Any]],
    ) -> Type[T]:
        """
        Injects field types into a model.
        """
        pass

    @abstractmethod
    def reduce_model_schema(
        self, model: Type[T], include_description: bool = True
    ) -> str:
        """
        Reduces the model schema into version with less tokens. Helpful for reducing prompt noise.
        """
        pass



from abc import ABC, abstractmethod
from typing import AsyncGenerator, Type, TypeVar

from pydantic import BaseModel

from ait.core.domain.interfaces import CompletionResponse

T = TypeVar("T", bound=BaseModel)


class LLMPort(ABC):
    """
    Abstract base class for LLM ports.
    """

    @abstractmethod
    async def chat(self, messages: list[dict[str, str]]) -> CompletionResponse:
        """
        Sends a message to the LLM and returns a text response.

        Args:
            messages (list[dict[str, str]]): The messages to generate a response to

        Returns:
            CompletionResponse: The response from the LLM
        """
        pass

    @abstractmethod
    async def stream(
        self, messages: list[dict[str, str]]
    ) -> AsyncGenerator[CompletionResponse, None]:
        """
        Sends a message to the LLM and streams the text response.

        Args:
            messages (list[dict[str, str]]): The messages to generate a response to

        Returns:
            AsyncGenerator[CompletionResponse, None]: The response from the LLM
        """
        pass

    @abstractmethod
    async def asend(
        self,
        messages: list[dict[str, str]],
        response_model: Type[T],
    ) -> CompletionResponse[T]:
        """
        Sends a message to the LLM asynchronously and returns an instance of the response model.

        Args:
            messages (list[dict[str, str]]): The messages to generate a response to
            response_model (Type[T]): The model to return the response as

        Returns:
            CompletionResponse[T]: The response from the LLM
        """
        pass



import os

from ait.adapters import InstructorAdapter, Jinja2Adapter, PydanticAdapter
from ait.core.ports import FormatterPort, LLMPort, ModellerPort


def create_llm_client(
    model: str,
    embedding_model: str,
    api_key: str,
    base_url: str | None = None,
) -> LLMPort:
    """
    Factory function to create an LLMClient instance with default configuration.

    Args:
        model (Optional[str]): The model to use for completions. Defaults to LLM_MODEL env var.
        embedding_model (Optional[str]): The model to use for embeddings. Defaults to EMBEDDING_MODEL env var.

    Returns:
        LLMClient: Configured LLM client instance

    Raises:
        ValueError: If required configuration is missing
    """
    model = model
    embedding_model = embedding_model or os.getenv(
        "EMBEDDING_MODEL", "text-embedding-ada-002"
    )
    return InstructorAdapter(
        model=model,
        embedding_model=embedding_model,
        api_key=api_key,
        base_url=base_url,
    )


def create_prompt_formatter() -> FormatterPort:
    """
    Factory function to create a PromptFormatter instance.

    Returns:
        PromptFormatter: Configured prompt formatter instance
    """
    return Jinja2Adapter()


def create_model_handler() -> ModellerPort:
    """
    Factory function to create a ResponseModelService instance.

    Returns:
        ResponseModelService: Configured model handler instance
    """
    return PydanticAdapter()



__version__ = "0.1.7"

from grafo import Chunk, Node, TreeExecutor

from .adapters import InstructorAdapter as LLMClient
from .adapters import Jinja2Adapter as PromptFormatter
from .adapters import PydanticAdapter as ModelService
from .core.base import BaseWorkflow
from .core.domain.errors import BaseError
from .core.domain.interfaces import CompletionResponse
from .core.tools import AITools

__all__ = [
    "LLMClient",
    "PromptFormatter",
    "ModelService",
    "AITools",
    "CompletionResponse",
    "Node",
    "TreeExecutor",
    "Chunk",
    "BaseWorkflow",
    "BaseError",
]



from abc import ABC, abstractmethod
from typing import Optional, Any


class FormatterPort(ABC):
    """Base class for LLM models."""

    @abstractmethod
    def render(self, path: str, input: Optional[dict[str, Any]] = None) -> str:
        """
        Render a template with variables.
        """
        pass



import os
from typing import Any, AsyncGenerator, Type, TypeVar

from pydantic import BaseModel
from toon_python import encode

from ait.core.domain.interfaces import CompletionResponse
from ait.factories import (
    create_llm_client,
    create_model_handler,
    create_prompt_formatter,
)

T = TypeVar("T", bound=BaseModel)
N = TypeVar("N")
C = TypeVar("C")


class AITools:
    """
    A class that bundles methods for easily interacting with LLMs and manipulating pydantic BaseModels.
    """

    def __init__(
        self,
        model: str | None = None,
        embedding_model: str | None = None,
        api_key: str | None = None,
        base_url: str | None = None,
    ):
        """
        Initialize AIT with LLM client and prompt formatter.

        Args:
            model (str): The model to use for completions
            embedding_model (str): The model to use for embeddings
            api_key (str): The API key for authentication
        """
        if not model:
            model = os.getenv("LLM_MODEL", "")
        if not embedding_model:
            embedding_model = os.getenv("EMBEDDING_MODEL", "")
        if not api_key:
            api_key = os.getenv("LLM_API_KEY", "")

        self.llm_client = create_llm_client(
            model=model,
            embedding_model=embedding_model,
            api_key=api_key,
            base_url=base_url,
        )
        self.prompt_formatter = create_prompt_formatter()
        self.model_handler = create_model_handler()

    def inject_types(
        self,
        model: Type[T],
        fields: list[tuple[str, Any]],
    ) -> Type[T]:
        """
        Injects field types into a response model.

        Args:
            model (Type[T]): The model to inject types into
            fields (list[tuple[str, Any]]): The fields to inject types into

        Returns:
            Type[T]: The model with injected types

        Example:
            >>> ait.inject_types(Fruit, [("name", Literal[tuple(available_fruits)])])
        """
        return self.model_handler.inject_types(model, fields)

    def reduce_model_schema(
        self, model: Type[T], include_description: bool = True
    ) -> str:
        """
        Reduces a response model schema into version with less tokens. Helpful for reducing prompt noise.

        Args:
            model (Type[T]): The model to reduce the schema of
            include_description (bool): Whether to include the description in the schema

        Returns:
            str: The reduced schema
        """
        return self.model_handler.reduce_model_schema(model, include_description)

    def _prepare_messages(self, path: str, **kwargs: Any) -> list:
        for key, value in kwargs.items():
            if isinstance(value, BaseModel):
                kwargs[key] = encode(value.model_dump_json())
            elif (
                isinstance(value, list)
                and len(value) > 0
                and isinstance(value[0], BaseModel)
                and all(isinstance(item, type(value[0])) for item in value)
            ):
                kwargs[key] = encode([item.model_dump_json() for item in value])
        prompt = self.prompt_formatter.render(
            path=path,
            input=kwargs,
        )
        return [
            {"role": "system", "content": prompt},
        ]

    async def chat(
        self,
        path: str,
        **kwargs: Any,
    ) -> CompletionResponse:
        """
        Execute a chat task and return a text response.

        Args:
            path (str): The path to the prompt template file
            **kwargs: Additional arguments to pass to the prompt formatter

        Returns:
            CompletionResponse: The response from the LLM with text content
        """
        messages = self._prepare_messages(path, **kwargs)
        return await self.llm_client.chat(messages=messages)

    async def stream(
        self,
        path: str,
        **kwargs: Any,
    ) -> AsyncGenerator[CompletionResponse, None]:
        """
        Execute a streaming task and return a stream of text responses.

        Args:
            path (str): The path to the prompt template file
            **kwargs: Additional arguments to pass to the prompt formatter

        Returns:
            AsyncGenerator[CompletionResponse, None]: Stream of responses from the LLM
        """
        messages = self._prepare_messages(path, **kwargs)
        async for response in self.llm_client.stream(messages=messages):
            yield response

    async def asend(
        self,
        response_model: Type[T],
        path: str,
        **kwargs: Any,
    ) -> CompletionResponse[T]:
        """
        Execute a structured task and return a typed response.

        Args:
            response_model (Type[T]): The model to return the response as
            path (str): Path to the prompt template file
            **kwargs: Additional arguments to pass to the prompt formatter

        Returns:
            CompletionResponse[T]: The response from the LLM with structured content
        """
        messages = self._prepare_messages(path, **kwargs)
        response = await self.llm_client.asend(
            messages=messages,
            response_model=response_model,
        )
        if not isinstance(response.content, response_model):
            raise ValueError(
                f"Response content is not an instance of {response_model.__name__}"
            )
        return response



from .instructor_adapter import InstructorAdapter
from .jinja2_adapter import Jinja2Adapter
from .pydantic_adapter import PydanticAdapter

__all__ = ["Jinja2Adapter", "PydanticAdapter", "InstructorAdapter"]



from http import HTTPStatus
from typing import Any, Type, TypeVar, Union

from grafo import Node
from pydantic import BaseModel

from ait.core.domain.errors import BaseError
from ait.core.ports import WorkflowPort
from ait.core.tools import AITools
from ait.core.utils import logger

S = TypeVar("S", bound=BaseModel)
V = TypeVar("V", bound=BaseModel)
T = TypeVar("T", bound=BaseModel)

MAX_RETRIES = 3


class BaseWorkflow(WorkflowPort):
    """
    Base class for agentic workflows.
    """

    def __init__(
        self,
        ait: AITools,
        error_class: Type[BaseError],
        max_retries: int = MAX_RETRIES,
        echo: bool = False,
    ):
        self.ait = ait
        self.ErrorClass = error_class
        self.echo = echo

        # Stateful context
        self.current_retries = 0
        self.max_retries = max_retries

    # * Methods
    async def task(
        self,
        path: str,
        response_model: Type[S] | None = None,
        **kwargs: Any,
    ) -> Union[str, S]:
        """
        Execute a task.

        Args:
            path (str): The path to the prompt template file
            response_model (Type[S] | None): The response model to return the response as
            **kwargs: Additional arguments to pass to the prompt formatter

        Returns:
            Union[str, S]: The response from the LLM with text content
        """

        if not response_model:
            response = await self.ait.chat(path=path, **kwargs)
            return response.content

        response = await self.ait.asend(
            response_model=response_model,
            path=path,
            **kwargs,
        )
        return response.content

    def _validate_output(self, node: Node[T], detail: str) -> T:
        """
        Validates the source output of a node.
        """
        if not node.output:
            raise self.ErrorClass(
                status_code=HTTPStatus.BAD_REQUEST.value,
                message=detail,
            )
        return node.output

    async def redirect(
        self,
        source_node: Node[S],
        validation_node: Node[V],
        target_nodes: list[Node[T]] | None = None,
    ):
        """
        Redirects the flow of the workflow based on the validation node output.

        Args:
            source_node (Node[S]): The source node.
            validation_node (Node[V]): The validation node.
            target_nodes (list[Node[T]], optional): The target nodes. Defaults to None.
        """
        source_output = self._validate_output(source_node, "Source node output is None")
        validation_output = self._validate_output(
            validation_node, "Validation node output is None"
        )
        is_valid = getattr(validation_output, "is_valid", False)
        self.current_retries += 1
        if self.current_retries > self.max_retries and not is_valid:
            for child in validation_node.children:
                await validation_node.disconnect(child)
            raise self.ErrorClass(
                status_code=HTTPStatus.BAD_REQUEST.value,
                message=f"Max retries reached. Validation node output: {validation_output.model_dump_json(indent=4)}",
            )

        if self.echo:
            logger.debug(f"Source Output: {source_output.model_dump_json(indent=2)}")
            logger.debug(
                f"Validation Output: {validation_output.model_dump_json(indent=2)}"
            )
        for child in validation_node.children:
            await validation_node.disconnect(child)

        if is_valid:
            if target_nodes:
                for target_node in target_nodes:
                    await validation_node.connect(target_node)
            return
        source_node.kwargs["eval"] = lambda: str(
            source_output.model_dump_json(indent=2)
        ) + str(getattr(validation_output, "reasoning", None))
        await validation_node.connect(source_node)

    async def run(self, *_: Any, **__: Any) -> Any:
        """
        Run the workflow.
        """
        raise NotImplementedError("Run method not implemented")



from typing import Generic, TypeVar

from openai.types.chat import ChatCompletion, ChatCompletionChunk
from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


class CompletionResponse(BaseModel, Generic[T]):
    """
    Data model for completion response.
    """

    completion: ChatCompletion | ChatCompletionChunk
    content: str | T

    @property
    def response_model(self) -> T:
        """
        Returns the instance of the response model of the completion response.
        """
        if isinstance(self.content, str) or isinstance(self.content, list):
            raise ValueError("Content is not structured.")
        return self.content



from httpx import HTTPError

from ait.core.utils import logger


class BaseError(HTTPError):
    """Base error class for all custom exceptions."""

    def __init__(
        self,
        status_code: int,
        message: str,
    ):
        self.status_code = status_code
        super().__init__(message=message)
        logger.error(
            f"\033[91m{self.__class__.__name__} ({self.status_code}): {str(message)}\033[0m"
        )


class LLMAdapterError(BaseError):
    """
    Exception raised when an error occurs in the LLM adapter.
    """



from .formatter_port import FormatterPort
from .llm_port import LLMPort
from .modeller_port import ModellerPort
from .workflow_port import WorkflowPort

__all__ = ["LLMPort", "FormatterPort", "ModellerPort", "WorkflowPort"]



from typing import Any, Optional

from jinja2 import Environment

from ait.core.ports import FormatterPort


class Jinja2Adapter(FormatterPort):
    """
    Jinja2 implementation of the formatter port.
    Supports only Markdown files.
    """

    def __init__(self):
        self.env = Environment()

    def _load_prompt(self, path: str) -> str:
        with open(path, "r", encoding="utf-8") as file:
            return file.read()

    def render(
        self,
        path: str,
        input: Optional[dict[str, Any]] = None,
    ) -> str:
        """
        Render a Markdown template from a path with variables.

        Args:
            path: The directory path where the file is located
            input: Variables to pass to the template
        """
        base_prompt = self._load_prompt(path)
        template = self.env.from_string(base_prompt)
        return template.render(**(input or {}))



import re
from typing import Any, Type, TypeVar

from pydantic import BaseModel, Field, create_model
from pydantic_core import PydanticUndefined

from ait.core.ports import ModellerPort

T = TypeVar("T", bound=BaseModel)


class PydanticAdapter(ModellerPort):
    """
    Service for creating Pydantic models from schemas.
    """

    def _normalize(self, text: str) -> str:
        """
        Normalizes the text to a valid Pydantic model field name.
        """
        return re.sub(r"[^a-zA-Z0-9_]", "", text)

    def _pascal_case(self, string: str) -> str:
        """
        Converts a string to pascal case.
        """
        normalized = re.sub(r"[^a-zA-Z0-9\s]", " ", string).strip()
        return "".join(word.capitalize() for word in normalized.split())

    def inject_types(
        self,
        model: Type[T],
        fields: list[tuple[str, Any]],
    ) -> Type[T]:
        """
        Injects field types into a model.
        """
        return create_model(
            model.__name__ + "Model",
            __base__=(model,),
            __doc__=model.__doc__,
            **{
                field_name: (
                    field_type,
                    Field(
                        description=model.model_fields[field_name].description,
                        examples=model.model_fields[field_name].examples,
                    ),
                )
                for (field_name, field_type) in fields
            },  # type: ignore
        )

    def reduce_model_schema(
        self, model: Type[T], include_description: bool = True
    ) -> str:
        """
        Reduces the model schema into version with less tokens. Helpful for reducing prompt noise.
        """
        reduced_schema = ""
        for field, info in model.model_fields.items():
            reduced_schema += (
                f"{field}({info.annotation}"
                + (
                    f", default={info.default})"
                    if info.default is not PydanticUndefined
                    else ")"
                )
                + (f": {info.description}" if include_description else "")
            )
        return reduced_schema
